@@ -3,10 +3,23 @@
 
 ((window) => {
   const core = window.Deno.core;
+  const {
+    ArrayIsArray,
+    ArrayPrototypeMap,
+    Error,
+    Uint8Array,
+    StringPrototypeStartsWith,
+    String,
+    SymbolIterator,
+  } = window.__bootstrap.primordials;
+  const webidl = window.__bootstrap.webidl;
+  const { URL } = window.__bootstrap.url;
   const { Window } = window.__bootstrap.globalInterfaces;
   const { getLocationHref } = window.__bootstrap.location;
   const { log, pathFromURL } = window.__bootstrap.util;
   const { defineEventHandler } = window.__bootstrap.webUtil;
+  const { deserializeJsMessageData, serializeJsMessageData } =
+    window.__bootstrap.messagePort;
 
   function createWorker(
     specifier,
@@ -34,8 +47,12 @@
     core.opSync("op_host_post_message", id, data);
   }
 
-  function hostGetMessage(id) {
-    return core.opAsync("op_host_get_message", id);
+  function hostRecvCtrl(id) {
+    return core.opAsync("op_host_recv_ctrl", id);
+  }
+
+  function hostRecvMessage(id) {
+    return core.opAsync("op_host_recv_message", id);
   }
 
   /**
@@ -68,13 +85,13 @@
           `Expected 'array' or 'boolean' for ${permission} permission, "${value}" received`,
         );
       }
-    } else if (!Array.isArray(value) && typeof value !== "boolean") {
+    } else if (!ArrayIsArray(value) && typeof value !== "boolean") {
       throw new Error(
         `Expected 'array' or 'boolean' for ${permission} permission, ${typeof value} received`,
       );
       //Casts URLs to absolute routes
-    } else if (Array.isArray(value)) {
-      value = value.map((route) => {
+    } else if (ArrayIsArray(value)) {
+      value = ArrayPrototypeMap(value, (route) => {
         if (route instanceof URL) {
           route = pathFromURL(route);
         }
@@ -167,8 +184,9 @@
       const sourceCode = core.decode(new Uint8Array());
 
       if (
-        specifier.startsWith("./") || specifier.startsWith("../") ||
-        specifier.startsWith("/") || type == "classic"
+        StringPrototypeStartsWith(specifier, "./") ||
+        StringPrototypeStartsWith(specifier, "../") ||
+        StringPrototypeStartsWith(specifier, "/") || type == "classic"
       ) {
         const baseUrl = getLocationHref();
         if (baseUrl != null) {
@@ -187,18 +205,9 @@
         options?.name,
       );
       this.#id = id;
-      this.#poll();
+      this.#pollControl();
+      this.#pollMessages();
     }
-
-    #handleMessage(data) {
-      const msgEvent = new MessageEvent("message", {
-        cancelable: false,
-        data,
-      });
-
-      this.dispatchEvent(msgEvent);
-    }
-
     #handleError(e) {
       const event = new ErrorEvent("error", {
         cancelable: true,
@@ -219,9 +228,9 @@
       return handled;
     }
 
-    #poll = async () => {
+    #pollControl = async () => {
       while (!this.#terminated) {
-        const [type, data] = await hostGetMessage(this.#id);
+        const [type, data] = await hostRecvCtrl(this.#id);
 
         // If terminate was called then we ignore all messages
         if (this.#terminated) {
@@ -229,11 +238,6 @@
         }
 
         switch (type) {
-          case 0: { // Message
-            const msg = core.deserialize(data);
-            this.#handleMessage(msg);
-            break;
-          }
           case 1: { // TerminalError
             this.#terminated = true;
           } /* falls through */
@@ -262,19 +266,57 @@
       }
     };
 
-    postMessage(message, transferOrOptions) {
-      if (transferOrOptions) {
-        throw new Error(
-          "Not yet implemented: `transfer` and `options` are not supported.",
+    #pollMessages = async () => {
+      while (!this.terminated) {
+        const data = await hostRecvMessage(this.#id);
+        if (data === null) break;
+        let message, transfer;
+        try {
+          const v = deserializeJsMessageData(data);
+          message = v[0];
+          transfer = v[1];
+        } catch (err) {
+          const event = new MessageEvent("messageerror", {
+            cancelable: false,
+            data: err,
+          });
+          this.dispatchEvent(event);
+          return;
+        }
+        const event = new MessageEvent("message", {
+          cancelable: false,
+          data: message,
+          ports: transfer,
+        });
+        this.dispatchEvent(event);
+      }
+    };
+
+    postMessage(message, transferOrOptions = {}) {
+      const prefix = "Failed to execute 'postMessage' on 'MessagePort'";
+      webidl.requiredArguments(arguments.length, 1, { prefix });
+      message = webidl.converters.any(message);
+      let options;
+      if (
+        webidl.type(transferOrOptions) === "Object" &&
+        transferOrOptions !== undefined &&
+        transferOrOptions[SymbolIterator] !== undefined
+      ) {
+        const transfer = webidl.converters["sequence<object>"](
+          transferOrOptions,
+          { prefix, context: "Argument 2" },
         );
+        options = { transfer };
+      } else {
+        options = webidl.converters.PostMessageOptions(transferOrOptions, {
+          prefix,
+          context: "Argument 2",
+        });
       }
-
-      if (this.#terminated) {
-        return;
-      }
-
-      const bufferMsg = core.serialize(message);
-      hostPostMessage(this.#id, bufferMsg);
+      const { transfer } = options;
+      const data = serializeJsMessageData(message, transfer);
+      if (this.#terminated) return;
+      hostPostMessage(this.#id, data);
     }
 
     terminate() {
