@@ -4,6 +4,8 @@ pub use deno_broadcast_channel;
 pub use deno_console;
 pub use deno_crypto;
 pub use deno_fetch;
+pub use deno_http;
+pub use deno_net;
 pub use deno_timers;
 pub use deno_url;
 pub use deno_web;
@@ -20,7 +22,6 @@ pub mod js;
 pub mod metrics;
 pub mod ops;
 pub mod permissions;
-pub mod resolve_addr;
 pub mod tokio_util;
 pub mod web_worker;
 pub mod worker;
