@@ -7,10 +7,9 @@ use crate::ops;
 use crate::permissions::Permissions;
 use deno_broadcast_channel::InMemoryBroadcastChannel;
 use deno_core::error::AnyError;
-use deno_core::error::Context as ErrorContext;
-use deno_core::futures::future::poll_fn;
 use deno_core::futures::stream::StreamExt;
 use deno_core::futures::Future;
+use deno_core::located_script_name;
 use deno_core::serde_json;
 use deno_core::serde_json::json;
 use deno_core::url::Url;
@@ -23,7 +22,8 @@ use deno_core::ModuleId;
 use deno_core::ModuleLoader;
 use deno_core::ModuleSpecifier;
 use deno_core::RuntimeOptions;
-use deno_web::BlobUrlStore;
+use deno_core::SharedArrayBufferStore;
+use deno_web::BlobStore;
 use log::debug;
 use std::env;
 use std::pin::Pin;
@@ -58,7 +58,6 @@ pub struct WorkerOptions {
   // of WebWorker
   pub create_web_worker_cb: Arc<ops::worker_host::CreateWebWorkerCb>,
   pub js_error_create_fn: Option<Rc<JsErrorCreateFn>>,
-  pub attach_inspector: bool,
   pub maybe_inspector_server: Option<Arc<InspectorServer>>,
   pub should_break_on_first_statement: bool,
   /// Sets `Deno.version.deno` in JS runtime.
@@ -70,8 +69,9 @@ pub struct WorkerOptions {
   pub get_error_class_fn: Option<GetErrorClassFn>,
   pub location: Option<Url>,
   pub origin_storage_dir: Option<std::path::PathBuf>,
-  pub blob_url_store: BlobUrlStore,
+  pub blob_store: BlobStore,
   pub broadcast_channel: InMemoryBroadcastChannel,
+  pub shared_array_buffer_store: Option<SharedArrayBufferStore>,
 }
 
 impl MainWorker {
@@ -96,10 +96,11 @@ impl MainWorker {
       deno_webidl::init(),
       deno_console::init(),
       deno_url::init(),
-      deno_web::init(options.blob_url_store.clone(), options.location.clone()),
+      deno_web::init(options.blob_store.clone(), options.location.clone()),
       deno_fetch::init::<Permissions>(
         options.user_agent.clone(),
         options.ca_data.clone(),
+        None,
       ),
       deno_websocket::init::<Permissions>(
         options.user_agent.clone(),
@@ -116,21 +117,21 @@ impl MainWorker {
       // Metrics
       metrics::init(),
       // Runtime ops
-      ops::runtime::init(main_module),
+      ops::runtime::init(main_module.clone()),
       ops::worker_host::init(options.create_web_worker_cb.clone()),
       ops::fs_events::init(),
       ops::fs::init(),
-      ops::http::init(),
       ops::io::init(),
       ops::io::init_stdio(),
-      ops::net::init(),
+      deno_net::init::<Permissions>(options.ca_data.clone(), options.unstable),
       ops::os::init(),
       ops::permissions::init(),
       ops::plugin::init(),
       ops::process::init(),
       ops::signal::init(),
-      ops::tls::init(),
       ops::tty::init(),
+      deno_http::init(),
+      ops::http::init(),
       // Permissions ext (worker specific state)
       perm_ext,
     ];
@@ -140,25 +141,25 @@ impl MainWorker {
       startup_snapshot: Some(js::deno_isolate_init()),
       js_error_create_fn: options.js_error_create_fn.clone(),
       get_error_class_fn: options.get_error_class_fn,
+      shared_array_buffer_store: options.shared_array_buffer_store.clone(),
       extensions,
-      attach_inspector: options.attach_inspector,
       ..Default::default()
     });
 
-    let mut should_break_on_first_statement = false;
-
-    if let Some(inspector) = js_runtime.inspector() {
-      if let Some(server) = options.maybe_inspector_server.clone() {
-        let session_sender = inspector.get_session_sender();
-        let deregister_rx = inspector.add_deregister_handler();
-        server.register_inspector(session_sender, deregister_rx);
-      }
-      should_break_on_first_statement = options.should_break_on_first_statement;
+    if let Some(server) = options.maybe_inspector_server.clone() {
+      let inspector = js_runtime.inspector();
+      let session_sender = inspector.get_session_sender();
+      let deregister_rx = inspector.add_deregister_handler();
+      server.register_inspector(
+        session_sender,
+        deregister_rx,
+        main_module.to_string(),
+      );
     }
 
     Self {
       js_runtime,
-      should_break_on_first_statement,
+      should_break_on_first_statement: options.should_break_on_first_statement,
     }
   }
 
@@ -183,17 +184,18 @@ impl MainWorker {
       serde_json::to_string_pretty(&runtime_options).unwrap()
     );
     self
-      .execute(&script)
+      .execute_script(&located_script_name!(), &script)
       .expect("Failed to execute bootstrap script");
   }
 
-  /// Same as execute2() but the filename defaults to "$CWD/__anonymous__".
-  pub fn execute(&mut self, js_source: &str) -> Result<(), AnyError> {
-    let path = env::current_dir()
-      .context("Failed to get current working directory")?
-      .join("__anonymous__");
-    let url = Url::from_file_path(path).unwrap();
-    self.js_runtime.execute(url.as_str(), js_source)
+  /// See [JsRuntime::execute_script](deno_core::JsRuntime::execute_script)
+  pub fn execute_script(
+    &mut self,
+    name: &str,
+    source_code: &str,
+  ) -> Result<(), AnyError> {
+    self.js_runtime.execute_script(name, source_code)?;
+    Ok(())
   }
 
   /// Loads and instantiates specified JavaScript module.
@@ -231,7 +233,6 @@ impl MainWorker {
       self
         .js_runtime
         .inspector()
-        .unwrap()
         .wait_for_session_and_break_on_next_statement()
     }
   }
@@ -239,7 +240,7 @@ impl MainWorker {
   /// Create new inspector session. This function panics if Worker
   /// was not configured to create inspector.
   pub async fn create_inspector_session(&mut self) -> LocalInspectorSession {
-    let inspector = self.js_runtime.inspector().unwrap();
+    let inspector = self.js_runtime.inspector();
     inspector.create_local_session()
   }
 
@@ -255,7 +256,7 @@ impl MainWorker {
     &mut self,
     wait_for_inspector: bool,
   ) -> Result<(), AnyError> {
-    poll_fn(|cx| self.poll_event_loop(cx, wait_for_inspector)).await
+    self.js_runtime.run_event_loop(wait_for_inspector).await
   }
 
   /// A utility function that runs provided future concurrently with the event loop.
@@ -295,7 +296,6 @@ mod tests {
       seed: None,
       js_error_create_fn: None,
       create_web_worker_cb: Arc::new(|_| unreachable!()),
-      attach_inspector: false,
       maybe_inspector_server: None,
       should_break_on_first_statement: false,
       module_loader: Rc::new(deno_core::FsModuleLoader),
@@ -305,8 +305,9 @@ mod tests {
       get_error_class_fn: None,
       location: None,
       origin_storage_dir: None,
-      blob_url_store: BlobUrlStore::default(),
+      blob_store: BlobStore::default(),
       broadcast_channel: InMemoryBroadcastChannel::default(),
+      shared_array_buffer_store: None,
     };
 
     MainWorker::from_options(main_module, permissions, &options)
